from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

from ..storage import (
    Int64 as int64,
    Double as double,
    AtomicInt64 as atomic_int64,
    Unlimited as unlimited,
    Weight as weight,
    Mean as mean,
    WeightedMean as weighted_mean,
)
